DEFAULT_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name})'
SHORT_FMT = '{asctime:<23s} [{levelname:<8s}] {message}'
PID_FMT = '{asctime:<23s} [{levelname:<8s}] {message} ({name} | PID-{process})'
RAW_FMT = '{message}'



"""Pandas utilities and partials of dataframe method calls.

Parameters that are known at program start are used to initialize the classes
so that, at runtime, dataframes can flow through a preconfigured processing
pipe of callable objects.

"""

from .read import ParquetReader
from .write import ParquetWriter
from .groupby import (
    FrameGroupByAgg,
    RollingGroupBy,
    RollingGroupByAgg,
    RollingGroupByApply
)
from .frame import (
    AsType,
    ColumnSelector,
    ColumnsSelector,
    ColumnMapper,
    RowsSelector,
    FrameGroupBy,
    Join,
    Assign,
    Drop,
    DropNA,
    SortValues,
    SetIndex,
    ResetIndex,
    Rename
)

__all__ = [
    'ParquetReader',
    'ParquetWriter',
    'AsType',
    'ColumnSelector',
    'ColumnsSelector',
    'ColumnMapper',
    'RowsSelector',
    'Join',
    'Assign',
    'Drop',
    'DropNA',
    'SortValues',
    'SetIndex',
    'ResetIndex',
    'Rename',
    'FrameGroupBy',
    'FrameGroupByAgg',
    'RollingGroupBy',
    'RollingGroupByAgg',
    'RollingGroupByApply'
]



from typing import Any
from collections.abc import Callable
from pathlib import Path
import tomli_w
import yaml
import json
import gzip
from ..misc import ArgRepr
from .misc import Toml, Yaml


class TomlWriter(ArgRepr):
    """Partial of the ``dump`` function in the ``tomli-w`` package.

    Parameters
    ----------
    path: str, optional
        Path (including file name) to save the TOML file to. May include any
        number of string placeholders (i.e., pairs of curly brackets) that will
        be interpolated when instances are called. Defaults to the current
        working directory of the python interpreter.
    overwrite: bool, optional
        What to do if the TOML file to write already exists.
        Defaults to ``False``.
    create: bool, optional
        What to do if the directory where the TOML file should be saved does
        not exist. Defaults to ``False``.
    prune: bool, optional
        Whether to silently drop non-string keys and ``None`` values from the
        dictionary-like object to save as TOML. Defaults to ``False```.
    multiline_strings: bool, optional
        Whether to preserve newline bytes in multiline strings.
        Defaults to ``False``.
    indent: int, optional
        Number of spaces for array indentation. Defaults to 4.
    **kwargs
        Additional keyword arguments will be forwarded python`s builtin
        ``open`` function. The `mode` is determined by the `overwrite`
        argument and is, therefore, dropped from the keyword arguments.

    """

    def __init__(
            self,
            path: str = '',
            overwrite: bool = False,
            create: bool = False,
            prune: bool = False,
            multiline_strings: bool = False,
            indent: int = 4,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.overwrite = overwrite
        self.create = create
        self.prune = prune
        self.multiline_strings = multiline_strings
        self.indent = indent
        self.kwargs = (kwargs.pop('mode', ''), kwargs)[1]
        super().__init__(
            self.path,
            overwrite,
            create,
            prune,
            multiline_strings,
            indent,
            **self.kwargs
        )

    @property
    def mode(self) -> str:
        """The mode to open the target file with."""
        return 'wb' if self.overwrite else 'xb'

    @staticmethod
    def __stop_recursion_for(obj: Any) -> bool:
        """Criterion to stop the recursion into a dictionary-like object."""
        try:
            mismatch = len([*obj]) != len(set(obj.keys()))
        except (AttributeError, TypeError):
            return True
        return mismatch or not hasattr(obj, '__getitem__')

    def _pruned(self, toml: Toml) -> Toml:
        """Recursively drop fields with ``None`` values and/or non-string keys.

        Parameters
        ----------
        toml
            The dictionary-like object to drop fields from.

        Returns
        -------
        dict
            Dictionary without non-string keys and ``None`` values.

        """
        if self.__stop_recursion_for(toml):
            return toml
        return {
            key: self._pruned(toml[key])
            for key in toml
            if isinstance(key, str) and toml[key] is not None
        }

    def __call__(self, toml: Toml, *parts: str) -> tuple[()]:
        """Serialize a dictionary-like object and write it to TOML file.

        Parameters
        ----------
        toml: dict
            The dictionary-like object to save as TOML.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        path = Path(self.path.format(*parts).strip())
        if self.create:
            path.parent.mkdir(parents=True, exist_ok=True)
        with path.open(self.mode, **self.kwargs) as file:
            tomli_w.dump(
                self._pruned(toml) if self.prune else toml,
                file,
                multiline_strings=self.multiline_strings,
                indent=self.indent
            )
        return ()


class YamlWriter(ArgRepr):
    """Partial of the ``dump`` function in the ``pyyaml`` package.

    Parameters
    ----------
    path: str, optional
        Path (including file name) to save the YAML file to. May include any
        number of string placeholders (i.e., pairs of curly brackets) that will
        be interpolated when instances are called. Defaults to the current
        working directory of the python interpreter.
    overwrite: bool, optional
        What to do if the YAML file to write already exists.
        Defaults to ``False``.
    create: bool, optional
        What to do if the directory where the YAML file should be saved does
        not exist. Defaults to ``False``.
    **kwargs
        Additional keyword arguments will be forwarded to the wrapped ``dump``
        method of the ``pyyaml`` package. See its `documentation <https://
        pyyaml.org/wiki/PyYAMLDocumentation>`__ for details.

    """

    def __init__(
            self,
            path: str = '',
            overwrite: bool = False,
            create: bool = False,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.overwrite = overwrite
        self.create = create
        self.kwargs = kwargs
        super().__init__(
            self.path,
            overwrite,
            create,
            **self.kwargs
        )

    @property
    def mode(self) -> str:
        """The mode to open the target file with."""
        return 'wt' if self.overwrite else 'xt'

    def __call__(self, yml: Yaml, *parts: str) -> tuple[()]:
        """Serialize the given object and write it to YAML file.

        Parameters
        ----------
        yml: dict or list
            The object to save as YAML.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        path = Path(self.path.format(*parts).strip())
        if self.create:
            path.parent.mkdir(parents=True, exist_ok=True)
        with path.open(self.mode) as file:
            yaml.dump(yml, file, **self.kwargs)
        return ()


class JsonWriter(ArgRepr):
    """Partial of the ``dump`` function in python's own ``json`` package.

    Parameters
    ----------
    path: str, optional
        Path (including file name) to save the JSON file to. May include any
        number of string placeholders (i.e., pairs of curly brackets) that will
        be interpolated when instances are called. Defaults to the current
        working directory of the python interpreter.
    overwrite: bool, optional
        What to do if the JSON file to write already exists.
        Defaults to ``False``.
    create: bool, optional
        What to do if the directory where the JSON file should be saved does
        not exist. Defaults to ``False``.
    gzipped: bool, optional
        Write the JSON to a gzip-compressed JSON file if ``True`` and a plain
        text file if ``False``. If left at ``None``, which is the default,
        file names ending in ".gz" will trigger compression whereas all other
        extensions (if any) will not.
    prune: bool, optional
        If ``True``, keys that are not of a basic type (``str``, ``int``,
        ``float``, ``bool``, ``None``) will be skipped instead of raising a
        ``TypeError``. Defaults to ``False``.
    default: callable, optional
        Called on objects that can’t otherwise be serialized. It should return
        a JSON encodable version of the object or raise a ``TypeError``.
        Defaults to ``None``, which raises a ``TypeError``.
    indent: int or str, optional
        If a positive integer or string, JSON array elements and object members
        will be pretty-printed with that indent level. A positive integer
        indents that many spaces per level. A string (such as "\t") is used to
        indent each level. If zero, negative, or "" (the empty string), only
        newlines are inserted. If ``None`` (the default), the most compact
        representation is used.
    **kwargs
        Additional keyword arguments will be forwarded to the builtin ``open``
        or the ``gzip.open`` function, depending on `gzipped`. The `mode` to
        open the file is hardcoded to with ``"wt"`` or ``"xt"`` (depending on
        `create`) and will, therefore, be dropped from the keyword arguments.

    """

    def __init__(
            self,
            path: str = '',
            overwrite: bool = False,
            create: bool = False,
            gzipped: bool | None = None,
            prune: bool = False,
            default: Callable[[Any], Any] | None = None,
            indent: int | str | None = None,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.overwrite = overwrite
        self.create = create
        self.gzipped = gzipped
        self.prune = prune
        self.default = default
        self.indent = indent
        self.kwargs = (kwargs.pop('mode', ''), kwargs)[1]
        super().__init__(
            self.path,
            overwrite,
            create,
            gzipped,
            prune,
            default,
            indent,
            **self.kwargs
        )

    @property
    def mode(self) -> str:
        """The mode to open the target file with."""
        return 'wt' if self.overwrite else 'xt'

    def __call__(self, obj: Yaml, *parts: str) -> tuple[()]:
        """Serialize the given object and write it to JSON file.

        Parameters
        ----------
        obj: dict or list
            The object to save as JSON.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        path = Path(self.path.format(*parts).strip())
        zipped = path.suffix == '.gz' if self.gzipped is None else self.gzipped
        if self.create:
            path.parent.mkdir(parents=True, exist_ok=True)
        with gzip.open(
                path, self.mode, **self.kwargs
        ) if zipped else path.open(
                self.mode, **self.kwargs
        ) as file:
            json.dump(
                obj,
                file,
                skipkeys=self.prune,
                ensure_ascii=False,
                indent=self.indent,
                default=self.default
            )
        return ()



from typing import Any
from collections.abc import Iterable, Sequence
from datetime import timedelta
from numpy import ndarray
from pandas import DataFrame as PandasFrame
from polars.dataframe.group_by import GroupBy as GroupByT, DynamicGroupBy
from polars import DataFrame, Series, Expr
from polars._typing import (
    IntoExprColumn,
    IntoExpr,
    ColumnNameOrSelector,
    StartBy,
    ClosedInterval,
    Label,
    JoinStrategy,
    JoinValidation,
    MaintainOrderJoin
)
from ..misc import ArgRepr
from .types import IntoExprs

type Column = ColumnNameOrSelector | Iterable[ColumnNameOrSelector]
type Predicate = (
    bool
    | list[bool]
    | ndarray[bool]
    | Series
    | IntoExprColumn
    | Iterable[IntoExprColumn]
)


class Select(ArgRepr):
    """Partial of the polars dataframe `select <select_>`__ method.

    Parameters
    ----------
    *exprs: IntoExpr
        Column(s) to select, specified as positional arguments. Accepts
        expression input. Strings are parsed as column names, other
        non-expression inputs are parsed as literals.
    **named_exprs: IntoExpr
        Additional columns to select, specified as keyword arguments. The
        columns will be renamed to the keyword used.


    .. _select: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
                polars.DataFrame.select.html

    """

    def __init__(self, *exprs: IntoExprs, **named_exprs: IntoExpr) -> None:
        super().__init__(*exprs, **named_exprs)
        self.exprs = exprs
        self.named_exprs = named_exprs

    def __call__(self, df: DataFrame) -> DataFrame:
        """Select columns from a polars dataframe.

        Parameters
        ----------
        df: DataFrame
            The Dataframe to select columns from.

        Returns
        -------
        DataFrame
            The selected columns.

        """
        return df.select(*self.exprs, **self.named_exprs)


class Filter(ArgRepr):
    """Partial of the polars dataframe `filter <filter_>`__ method.

    Parameters
    ----------
    *predicates:
        Expression(s) that evaluate to a boolean Series.
    **constraints
        Filter column(s) given named by the keyword argument itself by the
        supplied value. Constraints will be implicitly combined with other
        filters with a logical and.


    .. _filter: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
                polars.DataFrame.filter.html

    """

    def __init__(self, *predicates: Predicate, **constraints: Any) -> None:
        super().__init__(*predicates, **constraints)
        self.predicates = predicates
        self.constraints = constraints

    def __call__(self, df: DataFrame) -> DataFrame:
        """Filter dataframe by `predicates` anf value `constraints`.

        Parameters
        ----------
        df: DataFrame
            The dataframe to filter.

        Returns
        -------
        DataFrame
            The filtered dataframe.

        """
        return df.filter(*self.predicates, **self.constraints)


class Drop(ArgRepr):
    """Partial of the polars dataframe `drop <drop_>`__ method.

    Parameters
    ----------
    *columns: ColumnNameOrSelector
        Names of the columns that should be removed from the dataframe.
        Accepts column selector input.
    strict: bool, optional
        Validate that all column names exist in the current schema, and throw
        an exception if any do not. Defaults to ``True``


    .. _drop: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
              polars.DataFrame.drop.html

    """

    def __init__(self, *columns: Column, strict: bool = True) -> None:
        super().__init__(*columns, strict=strict)
        self.columns = columns
        self.strict = strict

    def __call__(self, df: DataFrame) -> DataFrame:
        """Drop `columns` from a polars dataframe.

        Parameters
        ----------
        df: DataFrame
            The Dataframe to drop columns from.

        Returns
        -------
        DataFrame
            The Dataframe without the dropped `columns`.

        """
        return df.drop(*self.columns, strict=self.strict)


class Sort(ArgRepr):
    """Partial of the polars dataframe `sort <sort_>`__ method.

    Parameters
    ----------
    by: IntoExpr
        Column(s) to sort by. Accepts expression input, including selectors.
        Strings are parsed as column names.
    *more_by: IntoExpr
        Additional columns to sort by, specified as positional arguments.
    descending: bool, optional
        Sort in descending order. When sorting by multiple columns, can be
        specified per column by passing a sequence of booleans.
        Defaults to ``False``.
    nulls_last: bool, optional
        Place null values last. Can be a single boolean applying to all
        columns or a sequence of booleans for per-column control.
        Defaults to ``False``
    multithreaded: bool, optional
        Sort using multiple threads. Defaults to ``True``.
    maintain_order: bool, optional
        Whether the order should be maintained if elements are equal.
        Defaults to ``False``.


    .. _sort: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
              polars.DataFrame.sort.html

    """

    def __init__(
            self,
            by: IntoExpr | Iterable[IntoExpr],
            *more_by: IntoExpr,
            descending: bool | Sequence[bool] = False,
            nulls_last: bool | Sequence[bool] = False,
            multithreaded: bool = True,
            maintain_order: bool = False,
    ) -> None:
        super().__init__(
            by,
            *more_by,
            descending=descending,
            nulls_last=nulls_last,
            multithreaded=multithreaded,
            maintain_order=maintain_order
        )
        self.by = by
        self.more_by = more_by
        self.descending = descending
        self.nulls_last = nulls_last
        self.multithreaded = multithreaded
        self.maintain_order = maintain_order

    def __call__(self, df: DataFrame) -> DataFrame:
        """Sort polars dataframe by column values.

        Parameters
        ----------
        df: DataFrame
            The Dataframe to sort.

        Returns
        -------
        DataFrame
            The sorted dataframe.

        """
        return df.sort(
            self.by,
            *self.more_by,
            descending=self.descending,
            nulls_last=self.nulls_last,
            multithreaded=self.multithreaded,
            maintain_order=self.maintain_order
        )


class WithColumns(ArgRepr):
    """Partial of the polars dataframe `with_columns <with_>`__ method.

    Parameters
    ----------
    *exprs: IntoExpr
        Column(s) to add, specified as positional arguments. Accepts expression
        input. Strings are parsed as column names, other non-expression inputs
        are parsed as literals.
    **named_exprs: IntoExpr
        Additional columns to add, specified as keyword arguments. The columns
        will be renamed to the keyword used.


    .. _with: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
              polars.DataFrame.with_columns.html

    """

    def __init__(self, *exprs: IntoExprs, **named_exprs: IntoExpr) -> None:
        super().__init__(*exprs, **named_exprs)
        self.exprs = exprs
        self.named_exprs = named_exprs

    def __call__(self, df: DataFrame) -> DataFrame:
        """Add or replace columns to/of a polars dataframe.

        Parameters
        ----------
        df: DataFrame
            The dataframe to add or replace columns to/of.

        Returns
        -------
        DataFrame
            The dataframe with columns added or replaced.

        """
        return df.with_columns(*self.exprs, **self.named_exprs)


class GroupBy(ArgRepr):
    """Partial of the polars dataframe `group_by <group_>`__ method.

    Parameters
    ----------
    *by: IntoExpr
        Column(s) to group by. Accepts expression input. Strings are parsed as
        column names.
    maintain_order: bool, optional
        Ensure that the order of the groups is consistent with the input data.
        This is slower than a default group by. Settings this to True blocks
        the possibility to run on the streaming engine. Default to ``False``.
    **named_by: IntoExpr
        Additional columns to group by, specified as keyword arguments.
        The columns will be renamed to the keyword used.


    .. _group: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
               polars.DataFrame.group_by.html

    """

    def __init__(
            self,
            *by: IntoExprs,
            maintain_order: bool = False,
            **named_by: IntoExpr,
    ) -> None:
        super().__init__(*by, maintain_order=maintain_order, **named_by)
        self.by = by
        self.maintain_order = maintain_order
        self.named_by = named_by

    def __call__(self, df: DataFrame) -> GroupByT:
        """Group a polars dataframe.

        Parameters
        ----------
        df: DataFrame
            The dataframe to group.

        Returns
        -------
        DataFrame
            The grouped dataframe.

        """
        return df.group_by(
            *self.by,
            maintain_order=self.maintain_order,
            **self.named_by
        )


class GroupByDynamic(ArgRepr):
    """Partial of the polars dataframe `group_by_dynamic <dyngrp_>`__ method.

    Parameters
    ----------
    index_column: IntoExpr
        Column used to group based on the time window. Often of type Date or
        Datetime. This column must be sorted in ascending order (or, if
        group_by is specified, then it must be sorted in ascending order
        within each group). In case of a dynamic group by on indices, dtype
        needs to be Int32 or Int64. Note that Int32 gets temporarily cast to
        Int64, so if performance matters use an Int64 column.
    every: str or timedelta
        Interval of the window. Suffix string of integer number with the
        letter "i" to indicate indexing by integer columns.
    period: str or timedelta, optional
        Length of the window. Equals ‘every’ if set to ``None`` (the default).
    offset: str or timedelta
        Offset of the window. Does not take effect if `start_by` is
        "datapoint". Defaults to zero.
    include_boundaries: bool, optional
        Add the lower and upper bound of the window to the "_lower_boundary"
        and “_upper_boundary” columns. This will impact performance because it
        is harder to parallelize. Defaults to ``False``.
    closed: "left", "right", "both", "none"
        Define which sides of the temporal interval are closed (inclusive).
    label: "left", "right", "datapoint"
        Which label to use for the window, lower boundary, upper boundary, or
        first value of the index column in the given window. If you don't need
        the label to be at one of the boundaries, choose this option for
        maximum performance.
    group_by: IntoExpr, optional
        Also group by this column/these columns. Defaults to ``None``.
    start_by: "window", "datapoint", "monday", "tuesday", ...
        The strategy to determine the start of the first window by, where
        "window" takes the earliest timestamp, truncates it with `every`, and
        then adds `offset`. Weekly windows start on Monday. "datapoint" starts
        from the first encountered data point, whereas any day of the week
        starts the window at the weekday before the first data point.
        The resulting window is then shifted back until the earliest datapoint
        is in or in front of it.


    .. _dyngrp: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
                polars.DataFrame.group_by_dynamic.html

    """

    def __init__(
            self,
            index_column: IntoExpr,
            every: str | timedelta,
            period: str | timedelta | None = None,
            offset: str | timedelta | None = None,
            include_boundaries: bool = False,
            closed: ClosedInterval = 'left',
            label: Label = 'left',
            group_by: IntoExprs | None = None,
            start_by: StartBy = 'window',
    ) -> None:
        self.index_column = index_column
        self.every = every
        self.period = period
        self.offset = offset
        self.include_boundaries = include_boundaries
        self.closed = closed.strip().lower()
        self.label = label.strip().lower()
        self.group_by = group_by
        self.start_by = start_by.strip().lower()
        super().__init__(
            index_column,
            every=every,
            period=period,
            offset=offset,
            closed=self.closed,
            label=self.label,
            group_by=group_by,
            start_by=self.start_by,
        )

    def __call__(self, df: DataFrame) -> DynamicGroupBy:
        """Evaluate rolling-window aggregations on a polars dataframe.

        Parameters
        ----------
        df: DataFrame
            The dataframe to compute rolling-window aggregations on.

        Returns
        -------
        DataFrame
            The rolling-window aggregations.

        """
        return df.group_by_dynamic(
            self.index_column,
            every=self.every,
            period=self.period,
            offset=self.offset,
            closed=self.closed,
            label=self.label,
            group_by=self.group_by,
            start_by=self.start_by
        )


class Join(ArgRepr):
    """Partial of the polars dataframe `join <join_>`__ method.

    Parameters
    ----------
    on: str
        Name(s) of the join columns in both DataFrames. If set, `left_on` and
        `right_on` should be ``None``. Should not be specified if `how` is
        "cross". Defaults to ``None``.
    how: "inner", "left", "right", "full", "semi", "anti", "cross"
        Join strategy.
    left_on: str, optional
        Name(s) of the left join column(s). Defaults to ``None``.
    right_on: str, optional
        Name(s) of the right join column(s). Defaults to ``None``.
    suffix: str, optional
        Suffix to append to columns with a duplicate name.
        Defaults to "_right".
    validate: "m:m", "m:1", "1:m", "1:1"
        Checks if join is of specified type, many-to-many, many-to-one,
        one-to_many, or one-to-one.
    nulls_equal: bool, optional
        Join on null values. By default, null values will never produce
        matches. Defaults to ``False``.
    coalesce: bool, optional
        Coalescing behavior (merging of join columns). Defaults to ``None``,
        which leaves the behaviour join specific.
    maintain_order: "none", "left", "right", "left_right", "right_left"
        Which dataframe row order to preserve, if any. Do not rely on any
        observed ordering without explicitly setting this parameter, as your
        code may break in a future release. Not specifying any ordering can
        improve performance Supported for inner, left, right and full joins.


    .. _join: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
              polars.DataFrame.join.html

    """

    def __init__(
            self,
            on: str | Expr | Sequence[str | Expr] | None = None,
            how: JoinStrategy = 'inner',
            left_on: str | Expr | Sequence[str | Expr] | None = None,
            right_on: str | Expr | Sequence[str | Expr] | None = None,
            suffix: str = '_right',
            validate: JoinValidation = 'm:m',
            nulls_equal: bool = False,
            coalesce: bool | None = None,
            maintain_order: MaintainOrderJoin | None = None,
    ) -> None:
        self.on = on
        self.how = how.strip().lower()
        self.left_on = left_on
        self.right_on = right_on
        self.suffix = suffix.strip()
        self.validate = validate.strip().lower()
        self.nulls_equal = nulls_equal
        self.coalesce = coalesce
        self.maintain_order = maintain_order
        super().__init__(
            on,
            self.how,
            left_on=left_on,
            right_on=right_on,
            suffix=self.suffix,
            validate=self.validate,
            nulls_equal=nulls_equal,
            coalesce=coalesce,
            maintain_order=maintain_order,
        )

    def __call__(self, left: DataFrame, right: DataFrame) -> DataFrame:
        """Join two polars dataframes.

        Parameters
        ----------
        left: DataFrame
            Left dataframe in the join.
        right: DataFrame
            Right dataframe in the join.

        Returns
        -------
        DataFrame
            The joined dataframes.

        """
        return left.join(
            right,
            self.on,
            self.how,
            left_on=self.left_on,
            right_on=self.right_on,
            suffix=self.suffix,
            validate=self.validate,
            nulls_equal=self.nulls_equal,
            coalesce=self.coalesce,
            maintain_order=self.maintain_order
        )


class ToPandas(ArgRepr):
    """Partial of the polars dataframe `to_pandas <pandas_>`__ method.

    Parameters
    ----------
    use_pyarrow_extension_array: bool, optional
        Use pyarrow-backed extension arrays instead of numpy arrays for the
        columns of the pandas dataframe. This allows zero copy operations and
        preservation of null values. Subsequent operations on the resulting
        pandas dataframe may trigger conversion to numpy if those operations
        are not supported by pyarrow compute. Defaults to ``False``.
    **kwargs
        Additional keyword arguments to be passed to `pyarrow.Table.to_pandas()
        <https://arrow.apache.org/docs/python/generated/pyarrow.Table.html#
        pyarrow.Table.to_pandas>`_.


    .. _pandas: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
                polars.DataFrame.to_pandas.html

    """

    def __init__(
            self,
            use_pyarrow_extension_array: bool = False,
            **kwargs: Any
    ) -> None:
        super().__init__(
            use_pyarrow_extension_array=use_pyarrow_extension_array,
            **kwargs
        )
        self.use_pyarrow_extension_array = use_pyarrow_extension_array
        self.kwargs = kwargs

    def __call__(self, df: DataFrame) -> PandasFrame:
        """Convert a polars dataframe into a pandas one.

        Parameters
        ----------
        df: DataFrame
            The polars dataframe to convert.

        Returns
        -------
        PandasFrame
            The converted pandas dataframe.

        """
        return df.to_pandas(
            use_pyarrow_extension_array=self.use_pyarrow_extension_array,
            **self.kwargs
        )



from typing import Any, Self
import torch.nn as ptn
from ..types import Tensor, Module, Resettable, Functional


class GatedEmbedder(Resettable):
    """Flexible Gated Linear Unit (GLU) for embedding a numerical feature.

    Parameters
    ----------
    mod_dim: int
        Desired embedding size. Will become the size of the last dimension of
        the output tensor.
    gate: Module or function, optional
        The activation function to be applied to half of the (linearly)
        projected input before multiplying with the other half. Must be
        a callable that accepts a tensor as sole argument, like a module from
        ``torch.nn`` or a function from ``torch.nn.functional``, depending
        on whether it needs to be further parameterized or not.
        Defaults to a sigmoid.
    inp_dim: int, optional
        The number of features to embed together. Defaults to 1.
    **kwargs
        Additional keyword arguments to pass through to the linear layer.

    """

    def __init__(
            self,
            mod_dim: int,
            gate: Module | Functional = ptn.Sigmoid(),
            inp_dim: int = 1,
            **kwargs: Any
    ) -> None:
        super().__init__()
        self.mod_dim = mod_dim
        # Although few, some activation functions have learnable parameters
        if hasattr(gate, 'reset_parameters'):
            gate.reset_parameters()
        self.gate = gate
        self.inp_dim = inp_dim
        self.kwargs = kwargs
        self.embed = ptn.Linear(inp_dim, 2 * mod_dim, **kwargs)

    def forward(self, inp: Tensor) -> Tensor:
        """Embed a single numerical feature through a Gated Linear Unit (GLU).

        Parameters
        ----------
        inp: Tensor
            The last dimension of the input tensor is typically expected to be
            of size 1 and to contain the numerical value of a single feature.
            In case `inp_dim` dim was explicitly set to a value > 1 on
            instantiation, the size of the last dimension must match `inp_dim`,
            the number of numerical features to embed together.

        Returns
        -------
        Tensor
            The output has the same number of dimensions as the input with
            the size of the last dimension changed to the specified `mod_dim`.

        """
        emb = self.embed(inp)
        return emb[..., :self.mod_dim] * self.gate(emb[..., self.mod_dim:])

    def reset_parameters(self) -> None:
        """Re-initialize all internal parameters."""
        self.embed.reset_parameters()
        # Although few, some activation functions have learnable parameters
        if hasattr(self.gate, 'reset_parameters'):
            self.gate.reset_parameters()

    def new(
            self,
            mod_dim: int | None = None,
            gate: Module | Functional | None = None,
            inp_dim: int | None = None,
            **kwargs: Any
    ) -> Self:
        """Return a fresh instance with the same or updated parameters.

        Parameters
        ----------
        mod_dim: int, optional
            Desired embedding size. Will become the size of the last dimension
            of the output tensor. Overwrites the `mod_dim` of the current
            instance if given. Defaults to ``None``.
        gate: Module or function, optional
            The activation function to be applied to half of the (linearly)
            projected input before multiplying with the other half. Must be
            a callable that accepts a tensor as sole argument, like a module
            from ``torch.nn`` or a function from ``torch.nn.functional``.
            Overwrites the `gate` of the current instance if given.
            Defaults to ``None``.
        inp_dim: int, optional
            The number of features to embed together. Overwrites the `inp_dim`
            of the current instance if given. Defaults to ``None``.
        **kwargs
            Additional keyword arguments are merged into the keyword arguments
            of the current instance and are then passed through to the linear
            layer together.

        Returns
        -------
        GatedEmbedder
            A fresh, new instance of itself.

        """
        return self.__class__(
            self.mod_dim if mod_dim is None else mod_dim,
            self.gate if gate is None else gate,
            self.inp_dim if inp_dim is None else inp_dim,
            **(self.kwargs | kwargs)
        )



from typing import Any, overload
import boto3
from botocore.client import BaseClient
from botocore.config import Config
from ...misc import ArgRepr


class S3(ArgRepr):
    """Wraps an S3 client for delayed instantiation and config encapsulation.

    For more convenient function chaining in functional compositions, instances
    are also callable and simply return a new client when called.

    Parameters
    ----------
    region_name: str, optional
        The name of the region associated with the client.
        Defaults to ``None``.
    api_version: str, optional
        The API version to use.  By default, botocore will use the latest API
        version when creating a client. You only need to specify this parameter
        if you want to use a previous API version of the client.
        Defaults to ``None``.
    use_ssl: bool, optional
        Whether to use SSL.  Defaults to ``True``.
    verify: bool or str, optional
        Whether to verify SSL certificates. Can be set to ``True``,
        a path/to/cert/bundle.pem, or ``False``. Defaults to ``True``.
    endpoint_url: str, optional
        The complete URL to use for the constructed client. Normally, botocore
        will automatically construct the appropriate URL to use when
        communicating with a service. You can specify a complete URL
        (including the "http/https" scheme) to override this behavior. If this
        value is provided, then `use_ssl` is ignored.
    aws_account_id: str, optional
        The account id to use when creating the client. Defaults to ``None``.
    aws_access_key_id: str, optional
        The access key to use when creating the client. Defaults to ``None``.
    aws_secret_access_key: str, optional
        The secret key to use when creating the client. Defaults to ``None``.
    aws_session_token: str, optional
        The session token to use for the client. Defaults to ``None``.
    **kwargs
        Additional parameters for the client. See the documentation of
        `context <https://boto3.amazonaws.com/v1/documentation/api/latest/
        guide/configuration.html#configuring-client-context-parameters>`_ and
        `config <https://botocore.amazonaws.com/v1/documentation/api/latest/
        reference/config.html>`_ for options. Defaults to ``None``.

    """

    def __init__(
            self,
            region_name: str | None = None,
            api_version: str | None = None,
            use_ssl: bool = True,
            verify: bool | str = True,
            endpoint_url: str | None = None,
            aws_account_id: str | None = None,
            aws_access_key_id: str | None = None,
            aws_secret_access_key: str | None = None,
            aws_session_token: str | None = None,
            **kwargs: Any
    ) -> None:
        self.region_name = self.__strip(region_name)
        self.api_version = self.__strip(api_version)
        self.use_ssl = use_ssl
        self.verify = verify.strip() if isinstance(verify, str) else verify
        self.endpoint_url = self.__strip(endpoint_url)
        self.__aws_account_id = aws_account_id
        self.__aws_access_key_id = aws_access_key_id
        self.__aws_secret_access_key = aws_secret_access_key
        self.__aws_session_token = aws_session_token
        self.kwargs = kwargs
        super().__init__(
            self.region_name,
            self.api_version,
            self.use_ssl,
            self.verify,
            self.endpoint_url,
            aws_account_id=self.__obfuscate(aws_account_id),
            aws_access_key_id=self.__obfuscate(aws_access_key_id),
            aws_secret_access_key=self.__obfuscate(aws_secret_access_key),
            aws_session_token=self.__obfuscate(aws_session_token),
            **self.kwargs
        )

    @staticmethod
    @overload
    def __strip(attr: None) -> None:
        ...

    @staticmethod
    @overload
    def __strip(attr: str) -> str:
        ...

    @staticmethod
    def __strip(attr):
        """Strip leading and trailing whitespaces from string arguments."""
        return attr if attr is None else attr.strip()

    @staticmethod
    @overload
    def __obfuscate(attr: None) -> None:
        ...

    @staticmethod
    @overload
    def __obfuscate(attr: str) -> str:
        ...

    @staticmethod
    def __obfuscate(attr):
        """Return mock string if secrets are not None."""
        return attr if attr is None else '****'

    @property
    def client(self) -> BaseClient:
        """A fresh, new  S3 client."""
        return boto3.client(
            service_name='s3',
            region_name=self.region_name,
            api_version=self.api_version,
            use_ssl=self.use_ssl,
            verify=self.verify,
            endpoint_url=self.endpoint_url,
            aws_account_id=self.__aws_account_id,
            aws_access_key_id=self.__aws_access_key_id,
            aws_secret_access_key=self.__aws_secret_access_key,
            aws_session_token=self.__aws_session_token,
            config=Config(**self.kwargs)
        )

    def __call__(self, *_: Any, **__: Any) -> BaseClient:
        """Return a fresh, new S3 client, ignoring any (keyword) arguments."""
        return self.client



from typing import overload
from polars._typing import SchemaDict
from ..misc import ArgRepr
from pandas import (
    Series as PandasSeries,
    DataFrame as PandasFrame,
    Index,
    DatetimeIndex
)
from polars import Series, DataFrame
import polars as pl


class FromPandas(ArgRepr):
    """Partial of the polars top-level function `from_pandas <pd_>`__.

    Parameters
    ----------
    schema_overrides: dict, optional
        Support override of inferred types for one or more columns.
        Defaults to ``None``.
    rechunk: bool, optional
        Make sure that all data is in contiguous memory. Default to ``True``.
    nan_to_null: bool, optional
        Pyarrow will convert the ``NaN`` to ``None``. Default to ``True``.
    include_index: bool, optional
        Load any non-default pandas indexes as columns. Default to ``False``.


    .. _pd: https://docs.pola.rs/api/python/stable/reference/api/
            polars.from_pandas.html

    """

    def __init__(
            self,
            schema_overrides: SchemaDict | None = None,
            rechunk: bool = True,
            nan_to_null: bool = True,
            include_index: bool = False,
    ) -> None:
        super().__init__(
            schema_overrides=schema_overrides,
            rechunk=rechunk,
            nan_to_null=nan_to_null,
            include_index=include_index
        )
        self.schema_overrides = schema_overrides
        self.rechunk = rechunk
        self.nan_to_null = nan_to_null
        self.include_index = include_index

    @overload
    def __call__(self, pandas: PandasSeries | Index | DatetimeIndex) -> Series:
        ...

    @overload
    def __call__(self, pandas: PandasFrame) -> DataFrame:
        ...

    def __call__(self, pandas):
        """Convert pandas structures into polars series or dataframes

        Parameters
        ----------
        pandas:
            Dataframe, series, or index to convert to polars.

        Returns
        -------
        Series or DataFrame
            Series if pandas series or index, dataframe otherwise.

        """
        return pl.from_pandas(
            pandas,
            schema_overrides=self.schema_overrides,
            rechunk=self.rechunk,
            nan_to_null=self.nan_to_null,
            include_index=self.include_index
        )



import warnings
import tomllib
from typing import Any
from collections.abc import Callable
from pathlib import Path
import gzip
import json
import yaml
from yaml import Loader
from ..misc import ArgRepr, NotFound, LiteralNotFound
from .misc import Toml, Yaml


class TomlReader(ArgRepr):
    """Light wrapper around the python standard library's ``tomllib.load``.

    Parameters
    ----------
    path: str, optional
        Directory under which the TOML file is located or full path to the
        TOML file. If not fully specified here, the path can be completed or
        changed on calling the instance. Defaults to the current working
        directory of the python interpreter.
    not_found: str, optional
        What to do if the specified TOML file is not found. One of "ignore",
        "warn", or "raise". Defaults to "raise". Use the ``NotFound`` enum to
        avoid typos!
    parse_float: callable, optional
        Will be called with the string of every TOML float to be decoded.
        Defaults to ``float``.
    **kwargs
        Additional keyword arguments will be forwarded python`s builtin
        ``open`` function. Since the `mode` must be ``"rb"``, it is dropped
        from the keyword arguments.

    See Also
    --------
    ~swak.misc.NotFound

    """

    def __init__(
            self,
            path: str = '',
            not_found: NotFound | LiteralNotFound = NotFound.RAISE,
            parse_float: Callable[[str], float] = float,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.not_found = not_found.strip().lower()
        self.parse_float = parse_float
        self.kwargs = (kwargs.pop('mode', ''), kwargs)[1]
        super().__init__(self.path, self.not_found, parse_float, **self.kwargs)

    def __call__(self, path: str = '') -> Toml:
        """Read a specific TOML file.

        If `not_found` is set to "warn" or "ignore" and the file cannot be
        found, an empty dictionary is returned.

        Parameters
        ----------
        path: str
            Path (including file name) to the TOML file to read. If it starts
            with a backslash, it will be interpreted as absolute, if not, as
            relative to the `path` specified at instantiation. Defaults to an
            empty string, which results in an unchanged `path`.

        Returns
        -------
        dict
            The parsed contents of the TOML file.

        """
        path = Path(self.path) / str(path).strip()
        try:
            with path.open('rb', **self.kwargs) as file:
                toml = tomllib.load(file, parse_float=self.parse_float)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty TOML.'
                    warnings.warn(msg.format(path.resolve()))
                    toml = {}
                case NotFound.IGNORE:
                    toml = {}
                case _:
                    raise error
        return toml


class YamlReader(ArgRepr):
    """Light wrapper around pyyaml's ``yaml.load`` function.

    Parameters
    ----------
    path: str, optional
        Directory under which the YAML file is located or full path to the
        YAML file. If not fully specified here, the path can be completed on
        calling the instance. Defaults to the current working directory of the
        python interpreter.
    not_found: str, optional
        What to do if the specified YAML file is not found. One of "ignore",
        "warn", or "raise". Defaults to "raise". Use the ``NotFound`` enum to
        avoid typos!
    loader: type, optional
        The loader class to use. Defaults to ``Loader``
    **kwargs
        Additional keyword arguments will be forwarded python`s builtin
        ``open`` function. The `mode` is hard-coded to ``"rb"`` and is dropped
        from the keyword arguments.

    See Also
    --------
    ~swak.misc.NotFound

    """

    def __init__(
            self,
            path: str = '',
            not_found: NotFound | LiteralNotFound = NotFound.RAISE,
            loader: type = Loader,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.not_found = not_found.strip().lower()
        self.loader = loader
        self.kwargs = (kwargs.pop('mode', ''), kwargs)[1]
        super().__init__(self.path, self.not_found, loader, **self.kwargs)

    def __call__(self, path: str = '') -> Yaml:
        """Read a specific YAML file.

        If `not_found` is set to "warn" or "ignore" and the file cannot be
        found, and empty dictionary is returned.

        Parameters
        ----------
        path: str, optional
            Path (including file name) to the YAML file to read. If it starts
            with a backslash, it will be interpreted as absolute, if not, as
            relative to the `path` specified at instantiation. Defaults to an
            empty string, which results in an unchanged `path`.

        Returns
        -------
        dict or list
            The parsed contents of the YAML file.

        """
        path = Path(self.path) / str(path).strip()
        try:
            with path.open('rb', **self.kwargs) as file:
                yml = yaml.load(file, self.loader)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty YAML.'
                    warnings.warn(msg.format(path.resolve()))
                    yml = {}
                case NotFound.IGNORE:
                    yml = {}
                case _:
                    raise error
        return yml or {}


class JsonReader(ArgRepr):
    """Light wrapper around python's own ``json.load`` function.

    Parameters
    ----------
    path: str, optional
        Directory under which the JSON file is located or full path to the
        JSON file. If not fully specified here, the path can be completed on
        calling the instance. Defaults to the current working directory of the
        python interpreter.
    not_found: str, optional
        What to do if the specified JSON file is not found. One of "ignore",
        "warn", or "raise". Defaults to "raise". Use the ``NotFound`` enum to
        avoid typos!
    gzipped: bool, optional
        Try to read a gzip-compressed JSON file if ``True`` and a plain text
        file if ``False``. If left at ``None``, which is the default, files
        will be interpreted as gzip-compressed if their name ends in ".gz"
        and as plain text if it does not.
    **kwargs
        Additional keyword arguments will be forwarded to the builtin ``open``
        or the ``gzip.open`` function, depending on `gzipped`. The `mode` to
        open the file is hardcoded to ``"rt"`` and will, therefore, be dropped
        from the keyword arguments.

    See Also
    --------
    ~swak.misc.NotFound

    """

    def __init__(
            self,
            path: str = '',
            not_found: NotFound | LiteralNotFound = NotFound.RAISE,
            gzipped: bool | None = None,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.not_found = not_found.strip().lower()
        self.gzipped = gzipped
        self.kwargs = (kwargs.pop('mode', ''), kwargs)[1]
        super().__init__(
            self.path,
            self.not_found,
            self.gzipped,
            **self.kwargs
        )

    def __call__(self, path: str = '') -> Yaml:
        """Read a specific JSON file.

        If `not_found` is set to "warn" or "ignore" and the file cannot be
        found, and empty dictionary is returned.

        Parameters
        ----------
        path: str, optional
            Path (including file name) to the JSON file to read. If it starts
            with a backslash, it will be interpreted as absolute, if not, as
            relative to the `path` specified at instantiation. Defaults to an
            empty string, which results in an unchanged `path`.

        Returns
        -------
        dict or list
            The parsed contents of the JSON file.

        """
        path = Path(self.path) / str(path).strip()
        zipped = path.suffix == '.gz' if self.gzipped is None else self.gzipped
        try:
            with gzip.open(
                    path, 'rt', **self.kwargs
            ) if zipped else path.open(
                    'rt', **self.kwargs
            ) as file:
                obj = json.load(file)
        except FileNotFoundError as error:
            match self.not_found:
                case NotFound.WARN:
                    msg = 'File {} not found!\nReturning empty JSON.'
                    warnings.warn(msg.format(path.resolve()))
                    obj = {}
                case NotFound.IGNORE:
                    obj = {}
                case _:
                    raise error
        return obj



from enum import StrEnum
from typing import Literal

type LiteralNotFound = Literal['ignore', 'warn', 'raise']
type LiteralBears = Literal['pandas', 'polars']


class NotFound(StrEnum):
    """Enum to direct read/load behaviour in case of missing files."""
    IGNORE = 'ignore'
    WARN = 'warn'
    RAISE = 'raise'


class Bears(StrEnum):
    """Enum to choose pandas versus polars."""
    PANDAS = 'pandas'
    POLARS = 'polars'



from typing import Literal, Any, overload
from collections.abc import Hashable, Callable, Sequence, Iterator, Mapping
from numpy import dtype, ndarray
from pandas.core.dtypes.base import ExtensionDtype
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
from pandas import DataFrame, Series, Index, Grouper
from ..misc.repr import ReprName
from ..misc import ArgRepr

type Type = str | type | dtype | ExtensionDtype
type Types = dict[Hashable, Type]
type Mask = list[bool] | Series | ndarray[bool]
type Condition = Callable[[DataFrame], Mask]
type Transform = dict[Hashable, Any] | Series | Callable[[Any], Any]
type Others = Series | DataFrame | list[Series | DataFrame]
type Key = Hashable | ndarray[Hashable] | Series | Index | Iterator[Hashable]
type Keys = list[Key]
type Renamer = Mapping[Hashable, Hashable] | Callable[[Hashable], Hashable]
type GroupKey = (
    str
    | Callable[[Hashable], Hashable]
    | Grouper
    | Mapping[Hashable, Hashable]
    | ndarray[Hashable]
    | Series
)
type GroupKeys = list[GroupKey]


class AsType(ReprName):
    """Partial of the ``pandas.DataFrame.astype`` method.

    Parameters
    ----------
    types: type or dict
        Single type or dictionary of column names and types, specifying type
        conversion of entire DataFrame or specific columns, respectively
    **kwargs
        Keyword arguments are passed on to the ``astype`` method call after
        the `types` argument.

    """
    def __init__(self, types: Type | Types, **kwargs: Any) -> None:
        super().__init__()
        self.types = types
        self.kwargs = kwargs

    def __repr__(self) -> str:
        kwargs = (f'{k}={self._repr(v)}' for k, v in self.kwargs.items())
        kwargs = ', '.join(kwargs)
        try:
            args = (f'{k!r}: {self._name(self.types[k])}' for k in self.types)
            args = '{' + ', '.join(args) + '}'
        except TypeError:
            args = self._name(self.types)
        signature = ', '.join(filter(None, [args, kwargs]))
        return f'{self.__class__.__name__}({signature})'

    def __call__(self, df: DataFrame) -> DataFrame:
        """Cast entire dataframe or columns thereof to specified types

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to type-cast.

        Returns
        -------
        DataFrame
            Pandas dataframe cast to new type or with columns cast to
            new types.

        """
        return df.astype(self.types, **self.kwargs)


class ColumnSelector(ArgRepr):
    """Select a single column of a (grouped) pandas dataframe as a series.

    This is simply a partial for calling a (grouped) dataframe's
    ``__getitem__`` method with a single argument (using the square-brackets
    accessor).

    Parameters
    ----------
    col: hashable
        Single DataFrame column to select.

    """

    def __init__(self, col: Hashable) -> None:
        self.col = self.__valid(col)
        super().__init__(col)

    @overload
    def __call__(self, df: DataFrame) -> Series:
        ...

    @overload
    def __call__(self, df: DataFrameGroupBy) -> SeriesGroupBy:
        ...

    def __call__(self, df):
        """Select a single column of a (grouped) pandas dataframe as series.

        Parameters
        ----------
        df: DataFrame or DataFrameGroupBy
            Pandas dataframe or grouped dataframe to select column from.

        Returns
        -------
        Series or SeriesGroupBy
            The selected column from the (grouped) dataframe.

        """
        return df[self.col]

    @staticmethod
    def __valid(col: Hashable) -> Hashable:
        _ = hash(col)
        return col


class ColumnsSelector(ArgRepr):
    """Select one or more columns of a (grouped) pandas dataframe as dataframe.

    This is simply a partial for calling a (grouped) dataframe's
    ``__getitem__``  method with a list of arguments (using the square-brackets
    accessor).

    Parameters
    ----------
    col: Hashable, optional
        Column name or sequence thereof. Defaults to an empty tuple.
    *cols: Hashable
        Additional columns names.

    """

    def __init__(
            self,
            col: Hashable | Sequence[Hashable] = (),
            *cols: Hashable
    ) -> None:
        col = self.__valid(col)
        self.cols: tuple[Hashable, ...] = col + self.__valid(cols)
        super().__init__(*self.cols)

    @overload
    def __call__(self, df: DataFrame) -> DataFrame:
        ...

    @overload
    def __call__(self, df: DataFrameGroupBy) -> DataFrameGroupBy:
        ...

    def __call__(self, df):
        """Select the specified column(s) from a (grouped) pandas dataframe.

        Parameters
        ----------
        df: DataFrame or DataFrameGroupBy
            Pandas dataframe or grouped dataframe to select column(s) from.

        Returns
        -------
        DataFrame or DataFrameGroupBy
            The selected column(s) of the (grouped) dataframe.

        """
        return df[list(self.cols)]

    @staticmethod
    def __valid(cols: Hashable | Sequence[Hashable]) -> tuple[Hashable, ...]:
        """Ensure that the columns are indeed a sequence of hashables."""
        if isinstance(cols, str):
            return cols,
        try:
            _ = [hash(col) for col in cols]
        except TypeError:
            _ = hash(cols)
            return cols,
        return tuple(cols)


class ColumnMapper(ArgRepr):
    """Transform one column of a pandas dataframe into another.

    This is simply a partial of calling the ``map`` method on one column of a
    dataframe and assigning the result to the same or another column of the
    same dataframe.

    Parameters
    ----------
    src_col: Hashable
        Column to call the ``map`` method on.
    transform: callable, Mapping, or Series
        Function or mapping in the form of a dictionary or a pandas series.
    tgt_col: Hashable, optional
        Dataframe column to store the series resulting from the
        transformation. Defaults to `src_col`, thus overwriting it in place.
    na_action: str, optional
        Can take the value "ignore" or ``None``, defaulting to the latter.
        Will be passed to the series ``map`` method along with `transform`.

    """

    def __init__(
            self,
            src_col: Hashable,
            transform: Transform,
            tgt_col: Hashable | None = None,
            na_action: Literal['ignore'] | None = None
    ) -> None:
        self.src_col = src_col
        self.tgt_col = src_col if tgt_col is None else tgt_col
        self.transform = transform
        self.na_action = na_action
        name = transform if callable(transform) else type(transform)
        super().__init__(src_col, name, tgt_col, na_action=na_action)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Called the ``map`` method on a specified column of a DataFrame.

        Cached keyword arguments are forwarded to the method call and
        the result is stored in the specified column of the DataFrame.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe with the column to call the ``map`` method on.

        Returns
        -------
        DataFrame
            Pandas dataframe with the result of the column transformation in
            the specified column.


        """
        df[self.tgt_col] = df[self.src_col].map(self.transform, self.na_action)
        return df


class RowsSelector(ArgRepr):
    """Select rows from a pandas dataframe with a boolean mask or function.

    This is simply a partial for calling a dataframe's ``__getitem__``
    method (using the square-brackets accessor) with a callable that takes
    the dataframe as input, and produces a 1-D, boolean array-like structure
    (of the same length as the dataframe to select from).

    Parameters
    ----------
    condition: callable or array-like
        A callable that accepts a dataframe and produces a 1-D, boolean array-
        like structure of the same length

    """

    def __init__(self, condition: Condition) -> None:
        super().__init__(condition)
        self.condition = condition

    def __call__(self, df: DataFrame) -> DataFrame:
        """Select rows from a dataframe with the specified mask or condition.

        Parameters
        ----------
        df: DataFrame
            The pandas dataframe to select rows from.

        Returns
        -------
        DataFrame
            The pandas dataframe with only the selected rows.

        """
        return df[self.condition]


class FrameGroupBy(ArgRepr):
    """Simple partial of a pandas dataframe's ``groupby`` method.

    Parameters
    ----------
    by: str, callable, series, array, dict, or list
        Column name, function (to be called on each column name), list or numpy
        array of the same length as the columns, a dict or series providing a
        label -> group name mapping, or a list of the above.
    level: hashable or sequence, optional
        If the axis is a multi-index (hierarchical), group by a particular
        level or levels. Do not specify both `by` and `level`.
        Defaults to ``None``.
    as_index: bool, optional
        Whether to return group labels as index. Defaults to ``True``.
    sort: bool, optional
        Whether to sort group keys. Defaults to ``True``.
    group_keys: bool, optional
        Defaults to ``True``
    observed: bool, optional
        Whether to show only observed values for categorical groupers.
        Defaults to ``False``.
    dropna: bool, optional
        Whether to treat NA values in group keys as groups.
        Defaults to ``True``.

    Note
    ----
    For a more extensive description of all (keyword) arguments, see the
    pandas `documentation <https://pandas.pydata.org/pandas-docs/stable/
    reference/api/pandas.DataFrame.groupby.html>`_.

    """

    def __init__(
            self,
            by: GroupKey | GroupKeys | None = None,
            level: Hashable | Sequence[Hashable] | None = None,
            as_index: bool = True,
            sort: bool = True,
            group_keys: bool = True,
            observed: bool = False,
            dropna: bool = True,
    ) -> None:
        super().__init__(
            by,
            level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna
        )
        self.by = by
        self.level = level
        self.as_index = as_index
        self.sort = sort
        self.group_keys = group_keys
        self.observed = observed
        self.dropna = dropna

    def __call__(self, df: DataFrame) -> DataFrameGroupBy:
        """Call the dataframe's ``groupby`` method with the cached (kw)args.

        Parameters
        ----------
        df: DataFrame
            Pandas dataframe to group.

        Returns
        -------
        DataFrameGroupBy
            The grouped dataframe.

        """
        return df.groupby(
            self.by,
            0,
            self.level,
            self.as_index,
            self.sort,
            self.group_keys,
            self.observed,
            self.dropna
        )


class Join(ArgRepr):
    """Light wrapper around the pandas dataframe ``join`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``join`` method call.
    **kwargs
        Keyword arguments to pass on to the ``join`` method call.

    Note
    ----
    For a full list of (keyword) arguments and their description, see the
    pandas `join documentation <https://pandas.pydata.org/pandas-docs/stable/
    reference/api/pandas.DataFrame.join.html>`_.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, df: DataFrame, other: Others) -> DataFrame:
        """Join a dataframe with other dataframe(s) and/or series.

        Parameters
        ----------
        df: DataFrame
            Source dataframe on which the ``join`` method will be called.
        other: DataFrame, Series, or a list of any combination
            Index should be similar to one (or more) columns in `df`. If a
            series is passed, its name attribute must be set, and that will be
            used as the column name in the resulting joined dataframe.

        Returns
        -------
        DataFrame
            The joined dataframe.

        """
        return df.join(other, *self.args, **self.kwargs)


class Assign(ArgRepr):
    """Light wrapper around a pandas dataframe's ``assign`` method.

    Parameters
    ----------
    col: dict
        A dictionary with the names of newly created (or overwritten) columns
        as keys. If the values are callable, they are computed on the entire
        dataframe and assigned to the new columns. The callable must not
        the change input dataframe (though pandas doesn’t check it). If the
        values are not callable, e.g., a series, scalar, or array, they are
        simply assigned.
    **cols
        As in the original, the keyword arguments themselves serve as the
        name(s) of the new (or overwritten) column(s) and their values are set
        in the same way.

    """

    def __init__(self, col: dict[str, Any] | None = None, **cols: Any) -> None:
        self.cols = cols if col is None else col | cols
        super().__init__(**self.cols)

    def __call__(self, df: DataFrame) -> DataFrame:
        """Add new columns to a dataframe by calling its ``assign`` method.

        Parameters
        ----------
        df: DataFrame
            The dataframe to add new columns to.

        Returns
        -------
        DataFrame
            The input dataframe with new columns added.

        """
        return df.assign(**self.cols)


class Drop(ArgRepr):
    """A simple partial of a dataframe's or series' ``drop`` method.

    Parameters
    ----------
    labels: hashable or sequence, optional
        Index or column labels to drop. Defaults to ``None``.
    axis: 1 or "columns", 0 or "index"
        Whether to drop labels from the columns (1 or "columns") or
        index (0 or "index"). Defaults to 1
    index: hashable or sequence, optional
        Single label or list-like. Defaults to ``None``.  Alternative to
        specifying axis (labels, axis=0 is equivalent to index=labels).
    columns: hashable or sequence, optional
        Single label or list-like. Defaults to ``None``. Alternative to
        specifying axis (labels, axis=1 is equivalent to columns=labels).
    level: hashable, optional.
        Integer or level name. Defaults to ``None``. For MultiIndex, level
        from which the labels will be removed.
    errors: "raise" or "ignore"
        Defaults to "raise". If "ignore", suppress error and drop only
        existing labels.

    """

    def __init__(
            self,
            label: Hashable | Sequence[Hashable] | None = None,
            *labels: Hashable,
            axis: int | Literal['index', 'columns', 'rows'] = 1,
            index: Hashable | Sequence[Hashable] | None = None,
            columns: Hashable | Sequence[Hashable] | None = None,
            level: Hashable | None = None,
            errors: Literal['ignore', 'raise'] = 'raise'
    ) -> None:
        self.labels = (self.__valid(label) + self.__valid(labels)) or None
        self.axis = axis
        self.index = index
        self.columns = columns
        self.level = level
        self.errors = errors
        super().__init__(
            self.labels,
            axis= axis,
            index=index,
            columns=columns,
            level=level,
            errors=errors
        )

    @overload
    def __call__(self, df: Series) -> Series:
        ...

    @overload
    def __call__(self, df: DataFrame) -> DataFrame:
        ...

    def __call__(self, df):
        """Drop rows or columns from a pandas series or dataframe.

        Parameters
        ----------
        df: Series or DataFrame
            The object to drop rows or columns from.

        Returns
        -------
        Series or DataFrame
            The object with rows or columns dropped.

        """
        return df.drop(
            self.labels,
            axis=self.axis,
            index=self.index,
            columns=self.columns,
            level=self.level,
            inplace=False,
            errors=self.errors,
        )

    @staticmethod
    def __valid(labels: Hashable | Sequence[Hashable]) -> list[Hashable]:
        """Ensure that the labels are indeed a sequence of hashables."""
        if labels is None:
            return []
        if isinstance(labels, str):
            return [labels]
        try:
            _ = [hash(label) for label in labels]
        except TypeError:
            _ = hash(labels)
            return [labels]
        return list(labels)


class DropNA(ArgRepr):
    """A simple partial of a dataframe's or series' ``dropna`` method.

    Parameters
    ----------
    axis: 0 or "index", 1 or "columns"
        Determine if rows or columns which contain missing values are removed.
        Defaults to 0.
    how: "any" or "all"
        Determine if row or column is removed from DataFrame, when we have at
        least one NA or all NA. Defaults to "any".
    thresh: int, optional
        Require that many non-NA values. Cannot be combined with how.
        Defaults to ``None``.
    subset: hashable or sequence, optional
        Labels along other axis to consider, e.g. if you are dropping rows
        these would be a list of columns to include. Defaults to ``None``.
    ignore_index: bool, optional
        Defaults to ``False``. If ``True``, the resulting axis will be labeled
        0, 1, …, n - 1.

    """

    def __init__(
            self,
            axis: int | Literal['index', 'columns', 'rows'] = 0,
            how: Literal['any', 'all'] | None = None,
            thresh: int | None = None,
            subset: Hashable | Sequence[Hashable] | None = None,
            ignore_index: bool = False
    ) -> None:
        self.axis = axis
        self.how = how
        self.thresh = thresh
        self.subset = subset
        self.ignore_index = ignore_index
        super().__init__(
            axis=self.axis,
            how=self.how,
            thresh=self.thresh,
            subset=self.subset,
            ignore_index=self.ignore_index
        )

    @overload
    def __call__(self, df: Series) -> Series:
        ...

    @overload
    def __call__(self, df: DataFrame) -> DataFrame:
        ...

    def __call__(self, df):
        """Drop rows or columns with NAs from a pandas series or dataframe.

        Parameters
        ----------
        df: Series or DataFrame
            The object to drop rows or columns with NAs from.

        Returns
        -------
        Series or DataFrame
            The object with rows or columns with NAs dropped.

        """
        return df.dropna(
            axis=self.axis,
            **({'how': self.how} if self.how else {'thresh': self.thresh}),
            subset=self.subset,
            inplace=False,
            ignore_index=self.ignore_index
        )


class SortValues(ArgRepr):
    """Partial of the pandas dataframe ``sort_values`` method.

    Parameters
    ----------
    by: hashable or sequence
        Name or list of names to sort by.
    **kwargs
        Additional keyword arguments will be forwarded to the method call with
        the exception of "inplace", which will be set to ``False``.

    Note
    ----
    For a full list of keyword arguments and their description, see the
    pandas `sort_values documentation <https://pandas.pydata.org/pandas-docs/
    stable/reference/api/pandas.DataFrame.sort_values.html>`_.

    """

    def __init__(
            self,
            by: Hashable | Sequence[Hashable],
            **kwargs: Any
    ) -> None:
        super().__init__(by, **kwargs)
        self.by = by
        self.kwargs = (kwargs.pop('inplace', ''), kwargs)[1]

    def __call__(self, df: DataFrame) -> DataFrame:
        """Sort a pandas dataframe by column(s) values.

        Parameters
        ----------
        df: DataFrame
            The dataframe to sort.

        Returns
        -------
        DataFrame
            The sorted dataframe.

        """
        return df.sort_values(self.by, inplace=False, **self.kwargs)


class SetIndex(ArgRepr):
    """Simple partial of a pandas dataframe's ``set_index`` method.

    Parameters
    ----------
    keys: hashable or array-like
        This parameter can be either a single column key, a single array of
        the same length as the calling DataFrame, or a list containing an
        arbitrary combination of column keys and arrays.
    drop : bool, optional
        Delete columns to be used as the new index. Defaults to ``True``.
    append : bool, optional
        Whether to append columns to existing index. Defaults to ``False``
    verify_integrity : bool, optional
        Whether to check the new index for duplicates. Defaults to ``False``.
        Setting to ``True`` will impact the performance of this method.

    """

    def __init__(
            self,
            keys: Key | Keys,
            drop: bool = True,
            append: bool = False,
            verify_integrity: bool = False
    ) -> None:
        self.keys = keys
        self.drop = drop
        self.append = append
        self.verify_integrity = verify_integrity
        super().__init__(
            keys,
            drop=drop,
            append=append,
            verify_integrity=verify_integrity
        )

    def __call__(self, df: DataFrame) -> DataFrame:
        """Set the index of a pandas dataframe.

        Parameters
        ----------
        df: DataFrame
            The dataframe to set the index of.

        Returns
        -------
        DataFrame
            The Dataframe with a new index set.

        """
        return df.set_index(
            self.keys,
            drop=self.drop,
            append=self.append,
            inplace=False,
            verify_integrity=self.verify_integrity
        )


class ResetIndex(ArgRepr):
    """Simple partial of a pandas dataframe's ``reset_index`` method.

    Parameters
    ----------
    level: int, str, tuple, or list, optional
        Only remove the given levels from the index. Defaults to ``None``,
        which removes all levels.
    drop: bool, optional
        Do not try to insert index into dataframe columns. This resets
        the index to the default integer index. Default to ``False``.
    col_level: int or str, optional
        If the columns have multiple levels, determines which level the
        labels are inserted into. Default to 0.
    col_fill: Hashable, optional
        If the columns have multiple levels, determines how the other
        levels are named. Defaults to an empty string.
    allow_duplicates : bool, optional
        Allow duplicate column labels to be created. Defaults to ``False``
    names : hashable or sequence, optional
        Using the given string, rename the dataframe column which contains the
        index data. If the dataframe has a multiindex, this has to be a list or
        tuple with length equal to the number of levels. Defaults to ``None``.

    """

    def __init__(
            self,
            level: Hashable | Sequence[Hashable] | None = None,
            drop: bool = False,
            col_level: Hashable = 0,
            col_fill: Hashable = '',
            allow_duplicates: bool = False,
            names: Hashable | Sequence[Hashable] | None = None,
    ) -> None:
        self.level = level
        self.drop = drop
        self.col_level = col_level
        self.col_fill = col_fill
        self.allow_duplicates = allow_duplicates
        self.names = names
        super().__init__(
            level,
            drop=drop,
            col_level=col_level,
            col_fill=col_fill,
            allow_duplicates=allow_duplicates,
            names=names
        )

    def __call__(self, df: DataFrame) -> DataFrame:
        """Reset the index of a pandas dataframe.

        Parameters
        ----------
        df: DataFrame
            The dataframe to reset the index of.

        Returns
        -------
        DataFrame
            The dataframe with its index reset.

        """
        return df.reset_index(
            self.level,
            drop=self.drop,
            inplace=False,
            col_level=self.col_level,
            col_fill=self.col_fill,
            allow_duplicates=self.allow_duplicates,
            names=self.names,
        )


class Rename(ArgRepr):
    """Simple partial of a pandas dataframe's ``rename`` method.

    Parameters
    ----------
    mapper : dict-like or function
        Dict-like or function transformations to apply to that axis' values.
    index : dict-like or function
        Alternative to specifying `mapper` with `axis` = 0.
    columns : dict-like or function
        Alternative to specifying `mapper` with `axis` = 1.
    axis : 1 or "columns", 0 or "index", optional
        Axis to target with `mapper`. Defaults to 1.
    level : Hashable, optional
        In case of a MultiIndex, only rename labels in the specified level.
        Defaults to ``None``
    errors : "ignore" or "raise", optional
        If "raise", raise a ``KeyError`` when a dict-like `mapper`, `index`,
        or `columns` contains labels that are not present in the index
        being transformed. If "ignore", existing keys will be renamed and
        extra keys will be ignored. Defaults to "ignore".

    """

    def __init__(
            self,
            mapper: Renamer | None = None,
            index: Renamer | None = None,
            columns: Renamer | None = None,
            axis: int | Literal['index', 'columns', 'rows'] = 1,
            level: Hashable | None = None,
            errors: Literal['ignore', 'raise'] = 'ignore'
    ) -> None:
        self.mapper = mapper
        self.index = index
        self.columns = columns
        self.axis = axis
        self.level = level
        self.errors = errors
        super().__init__(
            mapper,
            index=self.index,
            columns=self.columns,
            axis=self.axis,
            level=self.level,
            errors=self.errors
        )

    def __call__(self, df: DataFrame) -> DataFrame:
        """Rename a pandas dataframe's columns or rows.

        Parameters
        ----------
        df: DataFrame
            The dataframe to rename columns or rows of.

        Returns
        -------
        DataFrame
            The dataframe with renamed columns or rows.


        """
        return df.rename(
            self.mapper,
            index=self.index,
            columns=self.columns,
            axis=self.axis,
            level=self.level,
            inplace=False,
            errors=self.errors
        )



import sys
import logging
from typing import ParamSpec
from collections.abc import Callable
from functools import cached_property
from logging import Logger, Formatter, StreamHandler, Handler
from ...misc import ArgRepr
from .formats import DEFAULT_FMT

P = ParamSpec('P')
type Message = str | Callable[P, str]


class PassThroughStdLogger(ArgRepr):
    """Pass-through Logger to stdout or stderr with a formatted StreamHandler.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    fmt: str, optional
        Format string for the log messages in ``str.format()`` format.
        Defaults to "{asctime:<23s} [{levelname:<8s}] {message} ({name})".
    stream: str, optional
        Must be one of "stdout" or "stderr". Defaults to "stdout".

    Raises
    ------
    TypeError
        If `stream` is not a string.
    ValueError
        If `stream` is neither "stdout" nor "stderr"

    Note
    ----
    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Warnings
    --------
    To avoid creating and adding a new Handler to the same stream every time
    the same Logger is requested, one of its existing StreamHandlers will be
    modified if there are any. A new one will be created and added only if
    there aren't. By consequence, requesting the same Logger multiple times
    with a different `level` and/or a different `fmt` might change that
    Logger globally.

    """
    def __init__(
            self,
            name: str,
            level: int = logging.DEBUG,
            fmt: str = DEFAULT_FMT,
            stream: str = 'stdout'
    ) -> None:
        self.name = name.strip()
        self.level = level
        self.fmt = fmt
        self.stream = self.__valid(stream)
        super().__init__(self.name, level, fmt, stream)

    @staticmethod
    def __valid(stream: str) -> str:
        """Ensure that the provided stream is one the permitted options."""
        if not isinstance(stream, str):
            cls = type(stream).__name__
            msg = f'stream must be a string, not {cls}!'
            raise TypeError(msg)
        stream = stream.strip().lower()
        if stream not in ('stdout', 'stderr'):
            msg = f'stream must be "stdout" or "stderr", not "{stream}"!'
            raise ValueError(msg)
        return stream

    class Log:
        """Return type of the logging methods.

        Note
        ----
        This class is not intended to ever be instantiated directly.

        See Also
        --------
        PassThroughStdLogger.debug
        PassThroughStdLogger.info
        PassThroughStdLogger.warning
        PassThroughStdLogger.error
        PassThroughStdLogger.critical

        """

        def __init__(
                self,
                parent: 'PassThroughStdLogger',
                level: int,
                msg: Message
        ) -> None:
            self.parent = parent
            self.level = level
            self.msg = msg

        def __call__(self, *args: P.args) -> P.args:
            """Log the cached message and pass through the input argument(s).

            Parameters
            ----------
            *args
                If `msg` is a string, these arguments are inconsequential,
                and it will simply be logged. If, however, the `msg` is
                callable, it will be called with these arguments and the
                return value will be logged instead.

            Returns
            -------
            object or tuple
                Single argument if called with only one, otherwise all of them.

            """
            msg = self.msg(*args) if callable(self.msg) else self.msg
            self.parent.logger.log(self.level, msg)
            return args[0] if len(args) == 1 else args

    def debug(self, message: Message) -> Log:
        """Log a DEBUG message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.DEBUG, message)

    def info(self, message: Message):
        """Log an INFO message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.INFO, message)

    def warning(self, message: Message):
        """Log a WARNING message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.WARNING, message)

    def error(self, message: Message):
        """Log an ERROR message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.ERROR, message)

    def critical(self, message: Message):
        """Log a CRITICAL message, optionally depending on call arguments.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The requested Logger with one StreamHandler configured to specs."""
        # Get Logger with the given name.
        logger = logging.getLogger(self.name)
        # Adjust its log level so that messages from the Handler get through.
        logger.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Get all StreamHandlers to the requested stream.
        hdls = self.__filtered(logger.handlers)
        # If there are any, get the first one. If not, make a new one ...
        hdl = hdls[0] if hdls else StreamHandler(getattr(sys, self.stream))
        # ... and configure it to specs.
        configured = self.__configured(hdl)
        # If it was a new one, add it to the logger ...
        if not hdls:
            logger.addHandler(configured)
        # ... and return the logger.
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__handles_the_same_stream, handlers))

    def __handles_the_same_stream(self, handler: Handler) -> bool:
        """Is the Handler a StreamHandler to the requested stream?"""
        if isinstance(handler, StreamHandler):
            return handler.stream is getattr(sys, self.stream)
        return False

    def __configured(self, handler: StreamHandler) -> StreamHandler:
        """Configure the selected StreamHandler."""
        # Set the level of the handler
        handler.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Set the configured formatter
        handler.setFormatter(self.__formatter)
        return handler

    @property
    def __formatter(self) -> Formatter:
        """LogRecord Formatter for the specified format."""
        formatter = Formatter(fmt=self.fmt, style='{')
        formatter.default_msec_format = '%s.%03d'
        return formatter



from typing import Any
from pathlib import Path
from pandas import DataFrame
from ..misc import ArgRepr


class ParquetWriter(ArgRepr):
    """Partial of the pandas dataframe ``to_parquet`` method.

    Parameters
    ----------
    path: str, optional
        Path (including file name) to save the parquet file to. May include any
        number of string placeholders (i.e., pairs of curly brackets) that will
        be interpolated when instances are called. Defaults to the current
        working directory of the python interpreter.
    create: bool, optional
        What to do if the directory where the parquet file should be saved does
        not exist. Defaults to ``False``.
    **kwargs
        Keyword arguments passed on to the ``to_parquet`` method call.

    """

    def __init__(
            self,
            path: str = '',
            create: bool = False,
            **kwargs: Any
    ) -> None:
        self.path = str(path).strip()
        self.create = create
        self.kwargs = kwargs
        super().__init__(self.path, create, **kwargs)

    def __call__(self, df: DataFrame, *parts: str) -> tuple[()]:
        """Write a pandas DataFrame to parquet file.

        Parameters
        ----------
        df: DataFrame
            The dataframe to write.
        *parts: str, optional
            Fragments that will be interpolated into the `path` string given at
            instantiation. Obviously, there must be at least as many as there
            are placeholders in the `path`.

        Returns
        -------
        tuple
            An empty tuple.

        """
        path = Path(self.path.format(*parts).strip())
        if self.create:
            path.parent.mkdir(parents=True, exist_ok=True)
        df.to_parquet(str(path.resolve()), **self.kwargs)
        return ()



"""Polars utilities and partials of dataframe method calls.

Parameters that are known at program start are used to initialize the classes
so that, at runtime, dataframes can flow through a preconfigured processing
pipe of callable objects.

"""

from .misc import FromPandas
from .groupby import GroupByAgg
from .frame import (
    Select,
    Filter,
    Drop,
    Sort,
    WithColumns,
    GroupBy,
    GroupByDynamic,
    Join,
    ToPandas
)

__all__ = [
    'FromPandas',
    'Select',
    'Filter',
    'Drop',
    'Sort',
    'WithColumns',
    'GroupBy',
    'GroupByDynamic',
    'Join',
    'ToPandas',
    'GroupByAgg'
]



import math
from abc import ABC, abstractmethod
from ..types import Batches

__all__ = [
    'TestDataBase',
    'TrainDataBase'
]


class TestDataBase(ABC):

    @property
    @abstractmethod
    def n(self) -> int:
        """The total number of data points."""
        ...

    @abstractmethod
    def sample(self, batch_size: int, max_n: int | None = None) -> Batches:
        """Interator over batches from a reproducible sample of your data.

        Needed to consistently evaluate (and report) the train and/or test
        error after every epoch of training. Every time `sample` is called with
        the same arguments, the exact same output should be returned so that
        train and test errors can be used to track training convergence.

        Parameters
        ----------
        batch_size: int
            Number of data points in (or `batch_size` of) the reproducible
            sample of your data. Every time `sample` is called with the same
            `size`, the exact same output should be returned so that train and
            test errors can be used to track training convergence.
        max_n: int, optional
            Total number of data points to return mini-batches for. Defaults to
            ``None``, resulting in all available data points being used.
            To save some time, however, you might want to not use all available
            training data points just for evaluating your model metrics.

        Returns
        -------
        tuple
            A tuple of input tensors to call your model with.
        Tensor
            The matching target values. Must have the same dimensions and sizes
            as the output of your model.

        Important
        ---------
        Even if your model only takes a single tensor as input, this method
        must return a tuple to cover the general case of your model taking
        more than one tensor as input. If this is not the case, simple return
        a 1-tuple of tensors.

        """
        ...


class TrainDataBase(TestDataBase):

    def adjust_n_for(
            self,
            batch_size: int,
            step_freq: int = 1,
            n: int | None = None
    ) -> int:
        """Number of data points reduced to be suitably integer-divisible.

        This is a helper method for users to implement the ``__call__`` method
        in the case of `step_freq` > 1. Taking only the returned number of
        data points guarantees that all batches have the same size and that
        there will be no "left-over" batches at the end of the epoch.

        Parameters
        ----------
        batch_size: int
            The desired number of data points in one batch.
        step_freq: int, optional
            In case this number is > 1, the optimizer will accumulate gradients
            for that many batches before taking a step. All batches should be
            of the same size in this case and there shouldn't be any
            "left-over" batches at the end of each epoch. Defaults to 1.
        n: int, optional
            In rare cases, it might be useful to pass in the number of data
            points to adjust rather than taking the number of data points
            returned by ``self.n``, which is the default.

        Returns
        -------
        int
            Reduced number of data points that is guaranteed to be integer
            divisible by the product of `batch_size` and `step_freq`.

        """
        actual_n = self.n if n is None else n
        if step_freq <= 1:
            return actual_n
        super_batch_size = step_freq * batch_size
        return super_batch_size * (actual_n // super_batch_size)

    def adjust_batches_for(
            self,
            batch_size: int,
            step_freq: int= 1,
            n: int | None = None
    ) -> int:
        """Number of batches reduced to be suitably integer-divisible.

        This is a helper method for users to implement the ``__call__`` method
        in the case of `step_freq` > 1. The returned number of batches
        is guaranteed to be integer-divisible by `step_freq` so that no batches
        are "left over" at the end of the epoch.

        Parameters
        ----------
        batch_size: int
            The desired number of data points in one batch.
        step_freq: int, optional
            In case this number is > 1, the optimizer will accumulate gradients
            for that many batches before taking a step. All batches should be
            of the same size in this case and there shouldn't be any
            "left-over" batches at the end of each epoch. Defaults to 1.
        n: int, optional
            In rare cases, it might be useful to pass in the number of data
            points to adjust rather than taking the number of data points
            returned by ``self.n``, which is the default.

        Returns
        -------
        int
            Reduced number of batches that is guaranteed to be integer
            divisible by `step_freq`.

        """
        adjusted_n = self.adjust_n_for(batch_size, step_freq, n)
        return math.ceil(adjusted_n / batch_size)

    @abstractmethod
    def __call__(
            self,
            batch_size: int,
            step_freq: int = 1,
            epoch: int = 0
    ) -> tuple[int | None, Batches]:
        """Return an iterator over the mini-batches your model is trained on.

        Parameters
        ----------
        batch_size: int
            The (maximum) number of data points in one batch.
        step_freq: int, optional
            In case this number is > 1, the optimizer will accumulate gradients
            for that many batches before taking a step. All batches should be
            of the same size in this case. and there shouldn't be any
            "left-over" batches at the end of each epoch. Defaults to 1.
        epoch: int, optional
            In rare cases, it might be useful to know the current epoch when
            deciding which batches to return. Should start with 1 in the first
            epoch and, therefore, defaults to 0.

        Returns
        -------
        n_batches: int
            Total number of batches the returned iterator will provide.
            If unknown for some reason, also ``None`` can be returned.
        batches: Iterator
            One element yielded by the iterator is a 2-tuple. The first element
            is again a tuple, containing the input tensor(s) to call your model
            with. The second element of the tuple is a single tensor with
            the matching target values. It must have the same dimensions and
            sizes as the output of your model.

        Important
        ---------
        Even if your model only takes a single tensor as input, the first
        element of one tuple yielded by the returned iterator must always be a
        tuple of tensors to cover the general case of your model taking more
        than one tensor as input. If this is not the case, simple make that a
        1-tuple of tensors!

        """
        ...



import logging
from pathlib import Path
from functools import cached_property
from logging import Logger, Formatter, FileHandler, Handler
from ..repr import ArgRepr
from .formats import SHORT_FMT


class FileLogger(ArgRepr):
    """Wrapped logger to file with at least one formatted FileHandler.

    Parameters
    ----------
    file: str
        Full path to the file to log to, including file extension. If the
        parent directory does not exist, an attempt will be made to create it.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    fmt: str, optional
        Format string for the log messages in ``str.format()`` format.
        Defaults to "{asctime:<23s} [{levelname:<8s}] {message}".
    mode: str, optional
        Mode to open the file. Defaults to "a".
    encoding: str, optional
        Encoding to use when opening the file. Defaults to "utf-8".
    delay: bool, optional
        Whether to delay opening the file until it is first written to.
        Defaults to ``True``.

    Raises
    ------
    FileExistsError
        If a different Logger already has a FileHandler to the very same file.
        Because the instantiation of the actual Logger is delayed until it is
        first needed (to facilitate usage in multiprocessing scenarios),
        raising the exception is also delayed.

    Note
    ----
    The `mode` and the `encoding` can not be changed on an existing log file.

    """
    def __init__(
            self,
            file: str,
            level: int = logging.DEBUG,
            fmt: str = SHORT_FMT,
            mode: str = 'a',
            encoding: str = 'utf-8',
            delay: bool = True,
    ) -> None:
        self.file = str(Path(file.strip()).resolve())
        self.level = level
        self.fmt = fmt
        self.mode = mode.strip()
        self.encoding = encoding.strip()
        self.delay = delay
        super().__init__(
            self.file,
            self.mode,
            level,
            fmt,
            self.encoding,
            self.delay
        )

    def log(self, level: int, message: str) -> tuple[()]:
        """Log a message at the given level.

        Parameters
        ----------
        level: int
            The level to log at.
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        self.logger.log(level, message)
        return ()

    def debug(self, message: str) -> tuple[()]:
        """Log a DEBUG message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.DEBUG, message)

    def info(self, message: str) -> tuple[()]:
        """Log an INFO message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.INFO, message)

    def warning(self, message: str) -> tuple[()]:
        """Log a WARNING message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.WARNING, message)

    def error(self, message: str) -> tuple[()]:
        """Log an ERROR message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.ERROR, message)

    def critical(self, message: str) -> tuple[()]:
        """Log a CRITICAL message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The requested Logger with one FileHandler configured to specs."""
        # Check if some other Logger already has a Handler onto the same file.
        if self.handler_exists:
            msg = f'A different logger already handles file "{self.file}"!'
            raise FileExistsError(msg)
        # Get Logger with the given name.
        logger = logging.getLogger(self.name)
        # Adjust its log level so that messages from the Handler get through.
        logger.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Get all Handlers to the requested file. There should be at most 1!
        handlers = tuple(filter(self.__handles_file, logger.handlers))
        # If there are any, ...
        if handlers:
            # ... get the first one, ...
            hdl = handlers[0]
            # ... set its delay attribute, and, ...
            hdl.delay = self.delay
            # ... if the file does not yet exist, ...
            if not Path(self.name).exists():
                # ... also mode and encoding.
                hdl.mode = self.mode
                hdl.encoding = self.encoding
        # If there aren't any handlers yet, make a new one and create directory
        else:
            Path(self.file).parent.mkdir(parents=True, exist_ok=True)
            hdl = FileHandler(self.file, self.mode, self.encoding, self.delay)
        # Configure the FileHandler to specs.
        configured = self.__configured(hdl)
        # If it was a new one, add it to the logger ...
        if not handlers:
            logger.addHandler(configured)
        # ... and return the logger.
        return logger

    @property
    def name(self) -> str:
        """The name given to the Logger based on the provided file name."""
        return self.file.replace('.', '_')

    @property
    def handler_exists(self) -> bool:
        """Does a Logger with a Handler of the specified file already exist?"""
        root = logging.getLogger()
        candidates = [root] + list(root.manager.loggerDict.values())
        loggers = filter(lambda obj: isinstance(obj, Logger), candidates)
        return any(
            any(self.__handles_file(handler) for handler in logger.handlers)
            for logger in loggers
            if logger.name != self.name
        )

    def __handles_file(self, handler: Handler) -> bool:
        """Is the Handler a FileHandler to the requested file?"""
        if isinstance(handler, FileHandler):
            return handler.baseFilename == self.file
        return False

    def __configured(self, handler: FileHandler) -> FileHandler:
        """Configure the selected FileHandler."""
        # Set the level of the handler
        handler.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Set the configured formatter
        handler.setFormatter(self.__formatter)
        return handler

    @property
    def __formatter(self) -> Formatter:
        """LogRecord Formatter for the specified format."""
        formatter = Formatter(fmt=self.fmt, style='{')
        formatter.default_msec_format = '%s.%03d'
        return formatter



from typing import Any, Literal
from collections.abc import Callable
from numpy import ndarray
from pandas import DataFrame, Series
from pandas.core.groupby import DataFrameGroupBy
from pandas.core.window import RollingGroupby
from ..misc import ArgRepr

type Func = Callable[[Series], float] | str
type Funcs = list[Func] | dict[str, Func] | dict[str, list[Func]]


class FrameGroupByAgg(ArgRepr):
    """Simple partial for calling a grouped dataframe's ``agg`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``groupby`` method call.
    **kwargs
        Keyword arguments to pass on to the ``groupby`` method call.

    Note
    ----
    See the pandas `agg docs <https://pandas.pydata.org/pandas-docs/
    stable/reference/api/pandas.core.groupby.DataFrameGroupBy.agg.html>`_ for
    a full list of (keyword) arguments and an extensive description of
    usage and configuration.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, grouped_df: DataFrameGroupBy) -> DataFrame:
        """Call a grouped dataframe`s ``agg`` method with the cached (kw)args.

        Parameters
        ----------
        grouped_df: DataFrameGroupBy
            The grouped dataframe to aggregate.

        Returns
        -------
        DataFrame
            The aggregation of the grouped dataframe.

        """
        return grouped_df.agg(*self.args, **self.kwargs)


class RollingGroupBy(ArgRepr):
    """Simple partial of for calling a grouped dataframe's ``rolling`` method.

    Parameters
    ----------
    *args
        Arguments to pass on to the ``rolling`` method call.
    **kwargs
        Keyword arguments to pass on to the ``rolling`` method call.

    Notes
    -----
    See the pandas `rolling docs <https://pandas.pydata.org/docs/dev/reference
    /api/pandas.core.groupby.DataFrameGroupBy.rolling.html>`_ for a full list
    of (keyword) arguments and an extensive description of usage.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def __call__(self, grouped_df: DataFrameGroupBy) -> RollingGroupby:
        """Call a grouped dataframe`s ``rolling`` method with cached (kw)args.

        Parameters
        ----------
        grouped_df: DataFrameGroupBy
            The grouped dataframe to call ``rolling`` on.

        Returns
        -------
        RollingGroupby
            The rolling-grouped dataframe.

        """
        return grouped_df.rolling(*self.args, **self.kwargs)


class RollingGroupByAgg(ArgRepr):
    """Simple partial for calling a rolling-grouped dataframe's ``agg`` method.

    Parameters
    ----------
    func: callable, str, list, or dict
        Function(s) to use for aggregating the data. If a function, must work
        when passed a Series. Also acceptable are a function name, a list of
        function names and a dictionary with columns names as keys and
        functions, function names, or lists thereof as values.
    *args
        Positional arguments to pass to `func`.
    **kwargs
         Keyword arguments to pass to `func`.

    """

    def __init__(self, func: Func | Funcs, *args: Any, **kwargs: Any) -> None:
        super().__init__(func, *args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, rolling_df: RollingGroupby) -> DataFrame:
        """Call a rolling-grouped dataframe`s ``agg`` method.

        Parameters
        ----------
        rolling_df: RollingGroupby
            The rolling-grouped dataframe to aggregate.

        Returns
        -------
        DataFrame
            The aggregation of the rolling-grouped dataframe.

        """
        return rolling_df.agg(self.func, *self.args, **self.kwargs)


class RollingGroupByApply(ArgRepr):
    """Partial for calling a rolling-grouped dataframe's ``apply`` method.

    Parameters
    ----------
    func: callable
        Must produce a single numerical value from a numpy ndarray if
        `raw` = ``True`` or a series if `raw` = ``False``.
        Can also accept a numba JIT function with `engine` = "numba" specified.
    raw: bool, optional
        Whether to pass a numpy ndarray or a pandas series to `func`.
        Defaults to ``False``.
    engine: str, optional
        Either "cython" or "numba". Defaults to ``None``
    engine_kws: dict, optional
        Configuration of the "numba" engine. Keys can be "nopython", "nogil",
        and "parallel", and values must be ``True`` or ``False``.
        Defaults to ``None``.
    *args
        Positional arguments to pass to `func`.
    **kwargs
         Keyword arguments to pass to `func`.

    """

    def __init__(
            self,
            func: Callable[[Series | ndarray], float],
            raw: bool = False,
            engine: Literal['cython', 'numba'] | None = None,
            engine_kws: dict[str, bool] | None = None,
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__(func, raw, engine, engine_kws, *args, **kwargs)
        self.func = func
        self.raw = raw
        self.engine = engine
        self.engine_kws = engine_kws
        self.args = args
        self.kwargs = kwargs

    def __call__(self, rolling_df: RollingGroupby) -> DataFrame:
        """Call a rolling-grouped dataframe`s ``apply`` method.

        Parameters
        ----------
        rolling_df: RollingGroupby
            The rolling-grouped dataframe to aggregate.

        Returns
        -------
        DataFrame
            The aggregation of the rolling-grouped dataframe.

        """
        return rolling_df.apply(
            self.func,
            self.raw,
            self.engine,
            self.engine_kws,
            self.args,
            self.kwargs
        )



from .formats import DEFAULT_FMT, PID_FMT, RAW_FMT, SHORT_FMT
from .std import StdLogger
from .file import FileLogger

__all__ = [
    'DEFAULT_FMT',
    'SHORT_FMT',
    'PID_FMT',
    'RAW_FMT',
    'StdLogger',
    'FileLogger'
]



"""Tools to interact with elements of the Amazon Web Services (AWS).

Specifically, data scientists tend to interact heavily with S3 object storage.

"""

from .s3 import S3
from .df2s3 import DataFrame2S3Parquet
from .s32df import S3Parquet2DataFrame

__all__ = [
    'S3',
    'DataFrame2S3Parquet',
    'S3Parquet2DataFrame'
]



from polars.dataframe.group_by import GroupBy, DynamicGroupBy
from polars import DataFrame
from polars._typing import IntoExpr
from ..misc import ArgRepr
from .types import IntoExprs


class GroupByAgg(ArgRepr):
    """Partial of a polars (dynamic) group-by object's  `agg <agg_>`__ method.

    Parameters
    ----------
    *aggs: IntoExpr
        Aggregations to compute for each group of the group by operation,
        specified as positional arguments. Accepts expression input.
        Strings are parsed as column names.
    **named_aggs: IntoExpr
        Additional aggregations, specified as keyword arguments. The resulting
        columns will be renamed to the keyword used.


    .. _agg: https://docs.pola.rs/api/python/stable/reference/dataframe/api/
             polars.dataframe.group_by.GroupBy.agg.html#polars.dataframe.
             group_by.GroupBy.agg

    """

    def __init__(self, *aggs: IntoExprs, **named_aggs: IntoExpr) -> None:
        super().__init__(*aggs, **named_aggs)
        self.aggs = aggs
        self.named_aggs = named_aggs

    def __call__(self, grouped: GroupBy | DynamicGroupBy) -> DataFrame:
        """Aggregate a polars (dynamic) group-by object.

        Parameters
        ----------
        grouped: GroupBy or DynamicGroupBy
            The polars (dynamic) group-by object to aggregate.

        Returns
        -------
        DataFrame
            The aggregated (dynamic) group-by object.

        """
        return grouped.agg(*self.aggs, **self.named_aggs)



import sys
import logging
from functools import cached_property
from logging import Logger, Formatter, StreamHandler, Handler
from ..repr import ArgRepr
from .formats import DEFAULT_FMT


class StdLogger(ArgRepr):
    """Wrapped logger to stdout or stderr with a formatted StreamHandler.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    fmt: str, optional
        Format string for the log messages in ``str.format()`` format.
        Defaults to "{asctime:<23s} [{levelname:<8s}] {message} ({name})".
    stream: str, optional
        Must be one of "stdout" or "stderr". Defaults to "stdout".

    Raises
    ------
    TypeError
        If `stream` is not a string.
    ValueError
        If `stream` is neither "stdout" nor "stderr"

    Note
    ----
    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Warnings
    --------
    To avoid creating and adding a new Handler to the same stream every time
    the same Logger is requested, one of its existing StreamHandlers will be
    modified if there are any. A new one will be created and added only if
    there aren't. By consequence, requesting the same Logger multiple times
    with a different `level` and/or a different `fmt` might change that
    Logger globally.

    """
    def __init__(
            self,
            name: str,
            level: int = logging.DEBUG,
            fmt: str = DEFAULT_FMT,
            stream: str = 'stdout'
    ) -> None:
        self.name = name.strip()
        self.level = level
        self.fmt = fmt
        self.stream = self.__valid(stream)
        super().__init__(self.name, level, fmt, stream)

    @staticmethod
    def __valid(stream: str) -> str:
        """Ensure that the provided stream is one the permitted options."""
        if not isinstance(stream, str):
            cls = type(stream).__name__
            msg = f'stream must be a string, not {cls}!'
            raise TypeError(msg)
        stream = stream.strip().lower()
        if stream not in ('stdout', 'stderr'):
            msg = f'stream must be "stdout" or "stderr", not "{stream}"!'
            raise ValueError(msg)
        return stream

    def log(self, level: int, message: str) -> tuple[()]:
        """Log a message at the given level.

        Parameters
        ----------
        level: int
            The level to log at.
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        self.logger.log(level, message)
        return ()

    def debug(self, message: str) -> tuple[()]:
        """Log a DEBUG message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.DEBUG, message)

    def info(self, message: str) -> tuple[()]:
        """Log an INFO message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.INFO, message)

    def warning(self, message: str) -> tuple[()]:
        """Log a WARNING message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.WARNING, message)

    def error(self, message: str) -> tuple[()]:
        """Log an ERROR message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.ERROR, message)

    def critical(self, message: str) -> tuple[()]:
        """Log a CRITICAL message.

        Parameters
        ----------
        message: str
            The message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The requested Logger with one StreamHandler configured to specs."""
        # Get Logger with the given name.
        logger = logging.getLogger(self.name)
        # Adjust its log level so that messages from the Handler get through.
        logger.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Get all StreamHandlers to the requested stream.
        hdls = self.__filtered(logger.handlers)
        # If there are any, get the first one. If not, make a new one ...
        hdl = hdls[0] if hdls else StreamHandler(getattr(sys, self.stream))
        # ... and configure it to specs.
        configured = self.__configured(hdl)
        # If it was a new one, add it to the logger ...
        if not hdls:
            logger.addHandler(configured)
        # ... and return the logger.
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__handles_the_same_stream, handlers))

    def __handles_the_same_stream(self, handler: Handler) -> bool:
        """Is the Handler a StreamHandler to the requested stream?"""
        if isinstance(handler, StreamHandler):
            return handler.stream is getattr(sys, self.stream)
        return False

    def __configured(self, handler: StreamHandler) -> StreamHandler:
        """Configure the selected StreamHandler."""
        # Set the level of the handler
        handler.setLevel(min(max(self.level, logging.DEBUG), logging.CRITICAL))
        # Set the configured formatter
        handler.setFormatter(self.__formatter)
        return handler

    @property
    def __formatter(self) -> Formatter:
        """LogRecord Formatter for the specified format."""
        formatter = Formatter(fmt=self.fmt, style='{')
        formatter.default_msec_format = '%s.%03d'
        return formatter
